@@ -1,6 +1,59 @@
 library(caTools)
 library(PCICt)
 
+tapply.fast <- function (X, INDEX, FUN = NULL, ..., simplify = TRUE) {
+  FUN <- if (!is.null(FUN))
+    match.fun(FUN)
+  
+  if(!is.factor(INDEX))
+    stop("INDEX must be a factor.")
+  
+  if (length(INDEX) != length(X))
+    stop("arguments must have same length")
+  
+  if (is.null(FUN))
+    return(INDEX)
+  
+  namelist <- levels(INDEX)
+  ans <- lapply(split(X, INDEX), FUN, ...)
+  
+  ans <- unlist(ans, recursive = FALSE)
+  names(ans) <- levels(INDEX)
+  return(ans)
+}
+
+valid.climdexInput <- function(x) {
+  errors <- c()
+
+  if(!all(sapply(c("tmax", "tmin", "tavg", "prec", "dates", "annual.factor", "monthly.factor"), function(y) { length(slot(x, y)) }) == length(x@tmax)))
+    errors <- c(errors, "Data fields, dates, and factors must all be of the same length")
+
+  ## Check that namask.mon and namask.ann have columns for each of the variables
+  if(sum(c("tmax", "tmin", "tavg", "prec") %in% names(x@namask.ann)) != 4)
+    errors <- c(errors, "NA mask for annual must contain data for tmax, tmin, tavg, and prec.")
+
+  if(sum(c("tmax", "tmin", "tavg", "prec") %in% names(x@namask.mon)) != 4)
+    errors <- c(errors, "NA mask for monthly must contain data for tmax, tmin, tavg, and prec.")
+
+  ## Check that running.pctile.base running.pctile.notbase, and pctile all contain the proper named data.
+  bs.pctile.names <- c("tmax10", "tmax90", "tmin10", "tmin90")
+  pctile.names <- c("precwet95", "precwet99")
+
+  if(sum(bs.pctile.names %in% names(x@running.pctile.base)) != length(bs.pctile.names))
+    errors <- c(errors, "running.pctile.base does not contain at least one of tmax10, tmax90, tmin10, and tmin90.")
+
+  if(sum(bs.pctile.names %in% names(x@running.pctile.notbase)) != length(bs.pctile.names))
+    errors <- c(errors, "running.pctile.notbase does not contain at least one of tmax10, tmax90, tmin10, and tmin90.")
+
+  if(sum(pctile.names %in% names(x@pctile)) != length(pctile.names))
+    errors <- c(errors, "pctile does not contain at least one of precwet95 and precwet99.")
+  
+  if(length(errors) == 0)
+    return(TRUE)
+  else
+    return(errors)
+}
+
 setClass("climdexInput",
          representation(tmax = "numeric",
                         tmin = "numeric",
@@ -14,10 +67,10 @@ setClass("climdexInput",
                         dates = "PCICt",
                         base.range = "PCICt",
                         annual.factor = "factor",
-                        monthly.factor = "factor")
+                        monthly.factor = "factor"),
+         validity=valid.climdexInput
          )
 
-
 ## Returns PCICt field or dies
 get.date.field <- function(input.data, cal, date.types) {
   valid.date.types <- sapply(date.types, function(x) { return(!inherits(try(input.data[,x$fields], silent=TRUE), "try-error")) })
@@ -48,7 +101,7 @@ get.years <- function(dates) {
 }
 
 get.jdays.replaced.feb29 <- function(dates) {
-  return(unlist(tapply(get.jdays(dates), get.years(dates), function(x) { if(length(x) == 366) { return(c(1:59, 59, 60:365)) } else { return(x) } })))
+  return(unlist(tapply.fast(get.jdays(dates), factor(get.years(dates)), function(x) { if(length(x) == 366) { return(c(1:59, 59, 60:365)) } else { return(x) } })))
 }
 
 get.bootstrap.set <- function(dates, bootstrap.range, win.size) {
@@ -70,31 +123,29 @@ zhang.bootstrap.qtile <- function(x, dates, qtiles, bootstrap.range, include.mas
   window <- floor(n / 2)
 
   dpy <- ifelse(is.null(attr(dates, "dpy")), 365, attr(dates, "dpy"))
-  years.all <- get.years(dates)
-  jdays.idx <- get.jdays.replaced.feb29(dates)
   inset <- get.bootstrap.set(dates, bootstrap.range, n)
+  nyears <- floor(sum(inset) / dpy)
+  
+  if(!is.null(include.mask))
+    x[include.mask] <- NA
 
   bs.data <- x[inset]
-  jdays <- jdays.idx[inset]
-  if(!is.null(include.mask))
-    include.mask <- include.mask[inset]
   
   ## This routine is written as described in Zhang et al, 2005 as referenced above.
-  years <- years.all[inset]
-  year.list <- unique(years[(window + 1):(length(years) - window)])
-  d <- sapply(year.list, function(year.to.omit) {
-    bs.data.temp <- bs.data
-    omit.index <- years == year.to.omit
-    return(sapply(year.list[year.list != year.to.omit], function(year.to.replace.with) {
-      bs.data.temp[omit.index] <- bs.data.temp[years == year.to.replace.with]
-      return(running.quantile(bs.data.temp, n, qtiles, dpy, include.mask))
+  yidx <- 1:nyears
+  d <- sapply(yidx, function(idx.to.omit) {
+    ## Index is computed assuming `window` leading days of data...
+    omit.index <- window + (1:dpy) + ((idx.to.omit - 1) * dpy)
+    return(sapply(yidx[yidx != idx.to.omit], function(idx.to.replace.with) {
+      replace.index <- window + (1:dpy) + ((idx.to.replace.with - 1) * dpy)
+      bs.data[omit.index] <- bs.data[replace.index]
+      return(running.quantile(bs.data, n, qtiles, dpy))
     }))
   } )
-  byrs <- length(year.list)
   
-  dim(d) <- c(dpy, length(qtiles), byrs - 1, byrs)
+  dim(d) <- c(dpy, length(qtiles), nyears - 1, nyears)
   d <- aperm(d, perm=c(1, 4, 3, 2))
-  ## new dims: dpy, byrs, byrs-1, length(quantiles)
+  ## new dims: dpy, nyears, nyears-1, length(quantiles)
 
   return(lapply(1:length(qtiles), function(x) { d[,,,x] }))
 }
@@ -103,33 +154,59 @@ zhang.running.qtile <- function(x, dates, dates.base, qtiles, bootstrap.range, i
   jdays.idx <- get.jdays.replaced.feb29(dates)
   inset <- get.bootstrap.set(dates.base, bootstrap.range, n)
   dpy <- ifelse(is.null(attr(dates, "dpy")), 365, attr(dates, "dpy"))
+  
+  if(!is.null(include.mask))
+    x[include.mask] <- NA
 
   bs.data <- x[inset]
-  if(!is.null(include.mask))
-    include.mask <- include.mask[inset]
 
-  d <- apply(running.quantile(bs.data, n, qtiles, dpy, include.mask), 2, function(x) { return(x[jdays.idx]) } )
+  d <- apply(running.quantile(bs.data, n, qtiles, dpy), 2, function(x) { return(x[jdays.idx]) } )
 
   row.names(d) <- NULL
   return(d)
 }
 
 get.na.mask <- function(x, f, threshold) {
-  return(c(1, NA)[1 + as.numeric(tapply(is.na(x), f, function(y) { return(sum(y) > threshold) } ))])
+  return(c(1, NA)[1 + as.numeric(tapply.fast(is.na(x), f, function(y) { return(sum(y) > threshold) } ))])
+}
+
+get.num.days.in.range <- function(x, date.range) {
+  return(sum(x >= date.range[1] & x <= date.range[2]))
+  
 }
 
 climdexInput.raw <- function(tmax, tmin, prec, tmax.dates, tmin.dates, prec.dates, base.range=c(1961, 1990), n=5) {
+  if(length(tmin) != length(tmin.dates))
+    stop("Length of tmin data and tmin dates do not match.")
+  
+  if(length(tmax) != length(tmax.dates))
+    stop("Length of tmax data and tmax dates do not match.")
+  
+  if(length(prec) != length(prec.dates))
+    stop("Length of prec data and prec dates do not match.")
+
+  if(!(length(base.range) == 2 && is.numeric(base.range)))
+    stop("Invalid base date range; expecting vector of 2 numeric years.")
+
+  if(!(inherits(tmax.dates, "PCICt") && inherits(tmin.dates, "PCICt") && inherits(prec.dates, "PCICt")))
+    stop("Dates must all be of class PCICt.")
+
+  if(n != 5)
+    warning("Use of n != 5 varies from the Climdex definition. Use at your own risk.")
+  
   cal <- attr(tmax.dates, "cal")
   last.day.of.year <- "12-31"
   if(!is.null(attr(tmax.dates, "months")))
     last.day.of.year <- paste("12", attr(tmax.dates, "months")[12], sep="-")
   bs.date.range <- as.PCICt(paste(base.range, c("01-01", last.day.of.year), sep="-"), cal=cal)
+
+  ## Check that the base range includes at least a month's worth of data for each variable.
+  if(!all(sapply(list(tmax.dates, tmin.dates, prec.dates), get.num.days.in.range, bs.date.range) > 30))
+    warning("There is less than a month of data for at least one of tmin, tmax, and prec. Consider revising your base range and/or check your input data.")
+
   bs.win.date.range <- get.bootstrap.windowed.range(bs.date.range, n)
-  print("bs.win.date.range computed")
   all.dates <- c(tmin.dates, tmax.dates, prec.dates, bs.win.date.range)
-  print("Concatenated dates...")
   date.range <- range(all.dates)
-  print("Date.range computed")
   year.range <- as.numeric(strftime(date.range, "%Y", tz="GMT"))
   new.date.range <- as.PCICt(paste(year.range, c("01-01", last.day.of.year), sep="-"), cal=cal)
   date.series <- seq(new.date.range[1], new.date.range[2], by="day")
@@ -144,7 +221,7 @@ climdexInput.raw <- function(tmax, tmin, prec, tmax.dates, tmin.dates, prec.date
 
   filled.list <- list(filled.tmax, filled.tmin, filled.tavg, filled.prec)
   filled.list.names <- c("tmax", "tmin", "tavg", "prec")
-  
+
   namask.ann <- do.call(data.frame, lapply(filled.list, get.na.mask, annual.factor, 15))
   colnames(namask.ann) <- filled.list.names
   
@@ -167,10 +244,16 @@ climdexInput.raw <- function(tmax, tmin, prec, tmax.dates, tmin.dates, prec.date
   return(new("climdexInput", tmax=filled.tmax, tmin=filled.tmin, tavg=filled.tavg, prec=filled.prec, namask.ann=namask.ann, namask.mon=namask.mon, running.pctile.base=bs.pctile.base, running.pctile.notbase=bs.pctile, pctile=pctile, dates=date.series, base.range=bs.date.range, annual.factor=annual.factor, monthly.factor=monthly.factor))
 }
 
-climdexInput.csv <- function(tmax.file, tmin.file, prec.file, data.columns=list(tmin="tmin", tmax="tmax", prec="prec"), base.range=c(1961, 1990), na.strings=NULL, cal="gregorian", date.types=NULL) {
+climdexInput.csv <- function(tmax.file, tmin.file, prec.file, data.columns=list(tmin="tmin", tmax="tmax", prec="prec"), base.range=c(1961, 1990), na.strings=NULL, cal="gregorian", date.types=NULL, n=5) {
+  if(sum(c("tmax", "tmin", "prec") %in% names(data.columns)) != 3)
+    stop("Must specify names of all data columns (tmin, tmax, prec).")
+
   if(missing(date.types))
     date.types <- list(list(fields=c("year", "jday"), format="%Y %j"),
                        list(fields=c("year", "month", "day"), format="%Y %m %d"))
+  else
+    if(any(!sapply(date.types, function(x) { return(sum(c("fields", "format") %in% names(x)) == 2 && is.character(x$fields) && is.character(x$format)) } )))
+      stop("Invalid date.types specified. See ?climdexInput.csv .")
   
   tmin.dat <- read.csv(tmin.file, na.strings=na.strings)
   tmax.dat <- read.csv(tmax.file, na.strings=na.strings)
@@ -181,9 +264,8 @@ climdexInput.csv <- function(tmax.file, tmin.file, prec.file, data.columns=list(
   tmax.dat <- tmax.dat[!is.na(tmax.dat[,4]),]
   prec.dat <- prec.dat[!is.na(prec.dat[,4]),]
   
-  if(!(data.columns$tmin %in% names(tmin.dat) & data.columns$tmax %in% names(tmax.dat) & data.columns$prec %in% names(prec.dat))) {
+  if(!(data.columns$tmin %in% names(tmin.dat) & data.columns$tmax %in% names(tmax.dat) & data.columns$prec %in% names(prec.dat)))
     stop("Data columns not found in data.")
-  }
   
   tmin.dates <- get.date.field(tmin.dat, cal, date.types)
   tmax.dates <- get.date.field(tmax.dat, cal, date.types)
@@ -212,16 +294,16 @@ climdex.tr <- function(ci) { return(number.days.op.threshold(ci@tmin, ci@annual.
 climdex.gsl <- function(ci) { return(growing.season.length(ci@tavg, ci@annual.factor) * ci@namask.ann$tavg) }
 
 ## TXx: Monthly. Exact match.
-climdex.txx <- function(ci) { return(tapply(ci@tmax, ci@monthly.factor, max) * ci@namask.mon$tmax) }
+climdex.txx <- function(ci) { return(tapply.fast(ci@tmax, ci@monthly.factor, max) * ci@namask.mon$tmax) }
 
 ## TNx: Monthly. Exact match.
-climdex.tnx <- function(ci) { return(tapply(ci@tmin, ci@monthly.factor, max) * ci@namask.mon$tmin) }
+climdex.tnx <- function(ci) { return(tapply.fast(ci@tmin, ci@monthly.factor, max) * ci@namask.mon$tmin) }
 
 ## TXn: Monthly. Exact match.
-climdex.txn <- function(ci) { return(tapply(ci@tmax, ci@monthly.factor, min) * ci@namask.mon$tmax) }
+climdex.txn <- function(ci) { return(tapply.fast(ci@tmax, ci@monthly.factor, min) * ci@namask.mon$tmax) }
 
 ## TNn: Monthly. Exact match.
-climdex.tnn <- function(ci) { return(tapply(ci@tmin, ci@monthly.factor, min) * ci@namask.mon$tmin) }
+climdex.tnn <- function(ci) { return(tapply.fast(ci@tmin, ci@monthly.factor, min) * ci@namask.mon$tmin) }
 
 ## TN10p: Monthly. Pattern matches, but still significant differences.
 ## Our implementation currently follows the example set by fclimdex for dealing with missing values, which is wrong; it biases results upwards when missing values are present.
@@ -265,7 +347,11 @@ climdex.r10mm <- function(ci) { return(number.days.op.threshold(ci@prec, ci@annu
 climdex.r20mm <- function(ci) { return(number.days.op.threshold(ci@prec, ci@annual.factor, 20, ">=") * ci@namask.ann$prec) }
 
 ## Rnnmm: Annual. Exact match.
-climdex.rnnmm <- function(ci, threshold) { return(number.days.op.threshold(ci@prec, ci@annual.factor, threshold, ">=") * ci@namask.ann$prec) }
+climdex.rnnmm <- function(ci, threshold) {
+  if(!is.numeric(threshold) || length(threshold) != 1) stop("Please specify a single numeric threshold value.");
+
+  return(number.days.op.threshold(ci@prec, ci@annual.factor, threshold, ">=") * ci@namask.ann$prec)
+}
 
 ## Both CDD and CWD in fclimdex do not record the length of consecutive days on transition to a missing value
 ## CDD: Annual. Exact match.
@@ -309,7 +395,7 @@ get.series.lengths.at.ends <- function(x) {
 ## FD, ID, SU, TR, R10mm, R20mm, Rnnmm
 number.days.op.threshold <- function(temp, date.factor, threshold, op="<") {
   stopifnot(is.numeric(c(temp, threshold)))
-  return(tapply(match.fun(op)(temp, threshold), date.factor, sum, na.rm=TRUE))
+  return(tapply.fast(match.fun(op)(temp, threshold), date.factor, sum, na.rm=TRUE))
 }
 
 ## GSL
@@ -318,7 +404,7 @@ number.days.op.threshold <- function(temp, date.factor, threshold, op="<") {
 ## NOTE: There is a difference of 1 between our output and fclimdex. See line 637; consider case where start and end day are same. Correct answer is 1 day GSL; their answer is 0 day.
 growing.season.length <- function(daily.mean.temp, date.factor,
                                   min.length=6, t.thresh=5) {
-  return(tapply(daily.mean.temp, date.factor, function(ts) {
+  return(tapply.fast(daily.mean.temp, date.factor, function(ts) {
     ts.len<- length(ts)
     ts.mid <- floor(ts.len / 2)
     gs.begin <- which(select.blocks.gt.length(ts[1:(ts.mid-1)] > t.thresh, min.length - 1))
@@ -353,12 +439,12 @@ percent.days.op.threshold <- function(temp, dates, date.factor, threshold.outsid
     byrs <- (years.base.range[2] - years.base.range[1] + 1)
     year.base.list <- years.base.range[1]:years.base.range[2]
 
-    d <- lapply(1:byrs, function(x) { yset <-  years.base == year.base.list[x]; sapply(1:(byrs - 1), function(y) { f(temp.base[yset], (base.thresholds[,x,y])[jdays.base[yset]]) } ) })
+    d <- lapply(1:byrs, function(x) { yset <-  which(years.base == year.base.list[x]); sapply(1:(byrs - 1), function(y) { f(temp.base[yset], base.thresholds[jdays.base[yset],x,y]) } ) })
     ## This repeats a bug (or at least, debatable decision) in fclimdex where they always divide by byrs - 1 even when there are NAs (missing values) in the thresholds
     dat[inset] <- unlist(lapply(d, apply, 1, function(x) { sum(as.numeric(x), na.rm=TRUE) } ) ) / (byrs - 1)
   }
   
-  return(tapply(dat, date.factor, function(x) { x.nona <- x[!is.na(x)]; if(!length(x.nona)) return(NA); return(sum(x.nona) / length(x.nona) * 100) } ))
+  return(tapply.fast(dat, date.factor, function(x) { x.nona <- x[!is.na(x)]; if(!length(x.nona)) return(NA); return(sum(x.nona) / length(x.nona) * 100) } ))
 }
 
 ## WSDI, CSDI
@@ -369,19 +455,19 @@ threshold.exceedance.duration.index <- function(daily.max.temp, date.factor, war
             min.length > 0)
 
   periods <- select.blocks.gt.length(match.fun(op)(daily.max.temp, warm.thresholds), min.length - 1)
-  return(tapply(periods, date.factor, sum))
+  return(tapply.fast(periods, date.factor, sum))
 }
 
 ## DTR
 ## Max and min temps are assumed to be same length
 mean.daily.temp.range <- function(daily.max.temp, daily.min.temp, date.factor) {
-  return(tapply(daily.max.temp - daily.min.temp, date.factor, mean))
+  return(tapply.fast(daily.max.temp - daily.min.temp, date.factor, mean))
 }
 
 ## Rx1day, Rx5day
 max.nday.consec.prec <- function(daily.prec, date.factor, ndays) {
   if(ndays == 1) {
-    return(tapply(daily.prec, date.factor, max))
+    return(tapply.fast(daily.prec, date.factor, max))
   } else {
     ## Ends of the data will be de-emphasized (padded with zero precip data); NAs replaced with 0
     new.series <- c(rep(0, floor(ndays / 2)), daily.prec, rep(0, floor(ndays / 2)))
@@ -389,36 +475,32 @@ max.nday.consec.prec <- function(daily.prec, date.factor, ndays) {
     prec.runsum <- runmean(new.series, k=ndays, endrule="trim") * ndays
     ## Uncommenting this introduces the bug in RX5day in fclimdex, making the results identical
     ##prec.runsum <- c(0, 0, prec.runsum[1:(length(prec.runsum) - floor(ndays / 2))])
-    return(tapply(prec.runsum, date.factor, max))
+    return(tapply.fast(prec.runsum, date.factor, max))
   }
 }
 
 ## SDII
 ## Period for computation of number of wet days shall be the entire range of the data supplied.
 simple.precipitation.intensity.index <- function(daily.prec, date.factor) {
-  return(tapply(daily.prec, date.factor, function(prec) { idx <- prec >= 1 & !is.na(prec); if(sum(idx) == 0) { return(0); } else { return(sum(prec[idx], na.rm=TRUE) / sum(idx)) } } ))
+  return(tapply.fast(daily.prec, date.factor, function(prec) { idx <- prec >= 1 & !is.na(prec); if(sum(idx) == 0) { return(0); } else { return(sum(prec[idx], na.rm=TRUE) / sum(idx)) } } ))
 }
 
 ## CDD, CWD
 max.length.spell <- function(daily.prec, date.factor, threshold, op) {
   bools <- match.fun(op)(daily.prec, threshold)
-  return(tapply(get.series.lengths.at.ends(bools), date.factor, function(x) { return(max(x)) } ))
+  return(tapply.fast(get.series.lengths.at.ends(bools), date.factor, function(x) { return(max(x)) } ))
 }
 
 ## R95pTOT, R99pTOT
 total.precip.op.threshold <- function(daily.prec, date.factor, threshold, op) {
   f <- match.fun(op)
-  return(tapply(1:length(daily.prec), date.factor, function(x) { return(sum(daily.prec[x[f(daily.prec[x], threshold)]], na.rm=TRUE)) } ))
+  return(tapply.fast(1:length(daily.prec), date.factor, function(x) { return(sum(daily.prec[x[f(daily.prec[x], threshold)]], na.rm=TRUE)) } ))
 }
 
 ## Returns an n-day running quantile for each day of data
 ## Data is assumed to be padded by floor(n/2) days on either end, and data is assumed to start on the (dpy - floor(n/2) + 1)'th day..
-running.quantile <- function(data, n, q, dpy, include.mask=NULL) {
-  ## Apply include mask
-  if(!is.null(include.mask))
-    data[include.mask] <- NA
-
-  ret <- .Call("running_quantile_windowed", data, n, q, dpy)
+running.quantile <- function(data, n, q, dpy) {
+  ret <- .Call("running_quantile_windowed", data, n, q, dpy, DUP=FALSE)
   dim(ret) <- c(length(q), dpy)
   ##browser()
   return(t(ret))
